require 'rails_helper'

module Hermes
  require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')
  
  MUTATE_PARAMS = {
    is_active: 'Boolean',
    request_status: 'String',
    attributed_housing_district: 'String',
    attributed_housing_district_postal_code: 'String',
    entity_status: 'String',
    dalo: 'Boolean',
    asked_mobility: 'Boolean',
    valid_mobility: 'Boolean',
    decider: 'String',
    mobility: 'Boolean',
    seeked_types: 'Array',
    entity_help_required: 'Boolean',
    contract_type: 'String',
    registration_number: 'String',
    entity_validated: 'Boolean',
    priority_points_indexes: 'Array',
    nud: 'String',
    id_counter: 'String',
    id_unique: 'String',
    id_unique_associate: 'String',
    internal_ref: 'String',
    forcing_duplicate: 'Boolean',
    electronical_renewal: 'Boolean',
    ec_law_ressources: 'String',
    is_long_delay_anormal: 'Boolean',
    date_calculator_indicator: 'DateTime',
    with_car_park: 'Boolean',
    widen_request: 'Boolean',
    housing_type_list: 'Array',
    expulsion_judgement_date: 'DateTime',
    other_reason: 'String',
    work_contract_type: 'Array',
    seeked_localisation_list: 'Array',
    lessor_name: 'String',
    lessor_siren: 'String',
    social_housing_entry_date: 'Date',
    social_housing_name: 'String',
    test_created_at: 'DateTime',
    sne_number: 'String',
    employee_id: 'String',
    employee_start_date: 'DateTime',
    entity_access_level: 'String',
    employee_unit: 'String',
    tab_state: 'String',
    motivation1: 'String',
    motivation2: 'String',
    motivation3: 'String',
    seeking_first_floor: 'Boolean',
    seeking_elevator: 'Boolean',
    seeked_max_rent: 'Integer',
    sne_creation_date: 'DateTime',
    precisions: 'String',
    has_multiple_attributions_under_conditions: 'Boolean',
    seeked_housing_categories: 'Array',
    seeked_epci_first_quartils: 'Array',
    seeked_quartil_belongings: 'Array',
    dependents_monthly_incomes: 'Hash',
    dependents_monthly_charges: 'Hash',
    submitted_to_cal_nb: 'Integer',
    sent_to_cal_nb: 'Integer',
    rejected_in_cal_nb: 'Integer',
    rejected_after_cal_nb: 'Integer',
    accepted_in_cal_nb: 'Integer',
    radiation_date: 'Date',
    radiation_motive: 'String',
    documents_required: 'Integer',
    documents_uploaded: 'Integer',
    is_house_owner: 'Boolean',
    workplace_insee_code: "String"
  }.freeze
  
  ACCEPTABLE_PARAMS = %w[
    request_status
    als_id
    attributed_housing_district
    attributed_housing_district_postal_code
    old_request_status
    request_status_updated_at
    old_entity_status
    updated_entity_status
    entity_status
    dalo
    asked_mobility
    valid_mobility
    decider
    submitted_to_cal_nb
    sent_to_cal_nb
    rejected_in_cal_nb
    rejected_after_cal_nb
    accepted_in_cal_nb
    mobility
    seeked_types
    entity_help_required
    contract_type
    registration_number
    applicant_ids
    entity_validated
    priority_points_indexes
    nud
    nb_of_requests_made
    reference_tax_revenues
    seeked_districts
    single_parent
    id_counter
    id_unique
    id_unique_associate
    internal_ref
    forcing_duplicate
    electronical_renewal
    ec_law_ressources
    is_long_delay_anormal
    date_calculator_indicator
    with_car_park
    widen_request
    housing_type_list
    expulsion_judgement_date
    other_reason
    work_contract_type
    seeked_district_postal_codes
    seeked_departments
    seeked_localisation_list
    dependents
    monthly_revenues
    typology_range
    entity_validated
    family_size
    coapplicants
    lessor_name
    lessor_siren
    social_housing_entry_date
    social_housing_name
    test_created_at
    sne_number
    employee_id
    employee_start_date
    entity_access_level
    employee_unit
    tab_state
    motivation1
    motivation2
    motivation3
    seeking_first_floor
    seeking_elevator
    seeked_max_rent
    sne_creation_date
    precisions
    has_multiple_attributions_under_conditions
    seeked_housing_categories
    min_surface
    main_applicant_id
    main_applicant
    requester_firstname
    requester_lastname
    requester_fullname
    requester_email
    entity_id
    coapplicant_ids
    coapplicants
    seeked_epci_first_quartils
    seeked_quartil_belongings
    dependents_monthly_incomes
    dependents_monthly_charges
    housing_offer_ids
    radiation_date
    radiation_motive
    documents_required
    documents_uploaded
    is_house_owner
    last_application_date
    has_priority
  ].freeze
  
  VALID_OPERATORS = %w[
    $in
    $nin
    $ne
    $gt
    $gte
    $lt
    $lte
    $eq
    $regex
    $all
  ].freeze
  
  describe Api::Dmo::HousingRequestsController, :type => :api do
    routes {Hermes::Engine.routes}
    include Devise::Test::ControllerHelpers
    include JwtAuthHelper
    
    let!(:responsible) {create(:user, roles: ['Administrator'])}
    let!(:user) {create(:user)}
    let!(:applicant) {create(:applicant)}
    
    context 'with project oriented roles' do
      let!(:demandeur) {create(:user, auth_roles: ['Demandeur'])}
      let!(:demandeur_applicant) {create(:applicant)}
      
      let!(:bailleur) {create(:user, auth_roles: ['Bailleur'])}
      let!(:entreprise) {create(:user, auth_roles: ['Entreprise'])}
      
      describe 'POST as demandeur' do
        before :each do
          authenticated_header(request, demandeur)
          post 'create', {
            data: {
              attributes: {
                main_applicant_id: demandeur_applicant.id.to_s,
              }
            }
          }
        end
        
        it 'returns status code 201' do
          expect(response).to have_http_status(201)
        end
        
        it 'returns valid json' do
          attributes = json['data']['attributes']
          expect(attributes['main_applicant_id']).to eq demandeur_applicant.id.to_s
          expect(demandeur_applicant.reload.main_user_id).to eq demandeur.id
          expect(demandeur.reload.main_applicant_id).to eq demandeur_applicant.id
        end
      end
      
      describe 'POST as entreprise with main_user_id' do
        before :each do
          authenticated_header(request, entreprise)
          post 'create', {
            data: {
              attributes: {
                main_applicant_id: demandeur_applicant.id.to_s,
                main_user_id: demandeur.id.to_s
              }
            }
          }
        end
        
        it 'returns status code 201' do
          expect(response).to have_http_status(201)
        end
        
        it 'returns valid json' do
          attributes = json['data']['attributes']
          expect(attributes['main_applicant_id']).to eq demandeur_applicant.id.to_s
          expect(demandeur_applicant.reload.main_user_id).to eq demandeur.id
          expect(demandeur.reload.main_applicant_id).to eq demandeur_applicant.id
        end
      end
      
      describe 'POST as entreprise' do
        before :each do
          authenticated_header(request, entreprise)
          post 'create', {
            data: {
              attributes: {
                main_applicant_id: demandeur_applicant.id.to_s
              }
            }
          }
        end
        
        it 'returns status code 201' do
          expect(response).to have_http_status(201)
        end
        
        it 'returns valid json' do
          attributes = json['data']['attributes']
          expect(attributes['main_applicant_id']).to eq demandeur_applicant.id.to_s
          expect(demandeur_applicant.reload.main_user_id).to eq nil
          expect(demandeur.reload.main_applicant_id).to eq nil
        end
      end
    end
    
    # specific actions
    context 'for scoring and eligibility_tests' do
      let!(:threshold1) do
        create(
          :threshold,
          type: 'PLA LM',
          thresholds_map: {
            'Paris et communes limitrophes': {
              'thresholds_list': [
                13_050,
                21_272,
                27_883,
                30_521,
                36_307,
                40_859
              ],
              'another_person': 4552
            },
            'IDF hors Paris et Communes limitrophes': {
              'thresholds_list': [
                13_050,
                21_272,
                25_569,
                28_075,
                33_238,
                37_401
              ],
              'another_person': 4166
            },
            'Autres régions': {
              'thresholds_list': [
                11_342,
                16_525,
                19_872,
                22_111,
                25_870,
                29_155
              ],
              'another_person': 3252
            }
          }
        )
      end
      
      let!(:account1) do
        create(
          :user,
          first_name: 'Martin',
          last_name: 'Durand',
          email: "martin.durand1@gmail.com"
        )
      end
      
      let!(:applicant1) do
        create(
          :applicant,
          user_id: account1.id,
          amount_year_minus_two: 11675,
          monthly_incomes: {'Salaire': 1345.45}
        )
      end
      
      let!(:request1) do
        create(
          :housing_request,
          is_active: true,
          als_id: '1',
          seeked_localisation_list: [
            {codeDepartement: '34'}
          ],
          main_applicant: applicant1
        )
      end
      
      let!(:account2) do
        create(
          :user,
          first_name: 'Martin',
          last_name: 'Durand',
          email: "martin.durand2@gmail.com"
        )
      end
      
      let!(:applicant2) do
        create(
          :applicant,
          user_id: account2.id,
          amount_year_minus_two: 11675,
          monthly_incomes: {'Salaire': 1345.45}
        )
      end
      
      let!(:request2) do
        create(
          :housing_request,
          is_active: true,
          als_id: '2',
          seeked_localisation_list: [
            {codeDepartement: '24'}
          ],
          main_applicant: applicant2
        )
      end
      
      # def create_list(factory, nb, **params)
      #   puts factory
      #   puts nb
      #   nb.times.map { create(factory, **params) }
      # end
      
      let!(:offers_batch1) do
        create_list(
          :housing_offer,
          3,
          rent_amount: 234,
          typology: 'T2',
          department: '34',
          commercialized: false,
          threshold_type: 'PLA LM',
          threshold_subtype: 'Paris et communes limitrophes'
        )
      end
      
      let!(:offers_batch2) do
        create_list(
          :housing_offer,
          1,
          rent_amount: 234,
          typology: 'T2',
          department: '34',
          commercialized: true,
          lessor_status: 'RETIRE',
          threshold_type: 'PLA LM',
          threshold_subtype: 'Paris et communes limitrophes'
        )
      end
      
      let!(:offers_batch3) do
        create_list(
          :housing_offer,
          2,
          rent_amount: 234,
          typology: 'T2',
          department: '34',
          commercialized: true,
          lessor_status: 'NOT RETIRE',
          threshold_type: 'PLA LM',
          threshold_subtype: 'Paris et communes limitrophes'
        )
      end
      
      let!(:offers_batch4) do
        create_list(
          :housing_offer,
          2,
          rent_amount: 234,
          typology: 'T2',
          department: '24',
          commercialized: false,
          threshold_type: 'PLA LM',
          threshold_subtype: 'Paris et communes limitrophes'
        )
      end
      
      let!(:offers_batch5) do
        create_list(
          :housing_offer,
          4,
          rent_amount: 234,
          typology: 'T2',
          department: '24',
          commercialized: true,
          lessor_status: 'RETIRE',
          threshold_type: 'PLA LM',
          threshold_subtype: 'Paris et communes limitrophes'
        )
      end
      
      let!(:offers_batch6) do
        create_list(
          :housing_offer,
          1,
          rent_amount: 234,
          typology: 'T2',
          department: '24',
          commercialized: true,
          lessor_status: 'NOT RETIRE',
          threshold_type: 'PLA LM',
          threshold_subtype: 'Paris et communes limitrophes'
        )
      end
      
      describe 'GET /clp_details with ids' do
        before :each do
          authenticated_header(request, responsible)
          get 'clp_details', als_ids: %w[0000000001 0000000002]
        end
        
        it 'should generate valid eligibility' do
          expect(Hermes::HousingOffer.count).to be 13
          expect(
            Hermes::HousingOffer.where(applicant1.reload.eligibility_query.json_to_h)
              .count
          ).to be 6
          expect(
            Hermes::HousingOffer.where(applicant2.reload.eligibility_query.json_to_h)
              .count
          ).to be 7
        end
        
        it 'should has correct number of requests' do
          expect(json['data'].length).to be 2
        end
        
        it 'should eligible_offers_nb_detailed present for each' do
          expect(
            json['data'].map do |elt|
              elt['attributes']['eligible_offers_nb_detailed']
            end
          ).to all(be_present)
        end
        
        it 'should has correct details numbers' do
        
        end
      end
    end
    
    # common CRUD
    describe 'POST /housing_requests' do
      describe 'Create housing_request with minimal valid datas' do
        before :each do
          authenticated_header(request, responsible)
          post 'create', {
            data: {
              attributes: {
                main_applicant_id: applicant.id.to_s
              }
            }
          }
        end
        it 'returns status code 201' do
          expect(response).to have_http_status(201)
        end
      end
      MUTATE_PARAMS.each do |param, type|
        describe 'post request with ' + param.to_s + ' (' + type.to_s + ') is valid' do
          let(:params) do
            attr = {}
            attr[param] = get_typed_value(type)
            attr
          end
          before :each do
            params['main_applicant_id'] = applicant.id.to_s
            authenticated_header(request, responsible)
            post 'create', {
              data: {
                attributes: params
              }
            }
          end
          it 'returns status code 201 and the value appears in json' do
            expect(response).to have_http_status(201)
            if (type == 'Date' || type == 'DateTime')
              expect(json['data']['attributes'][param.to_s]).not_to be_empty
              expect(json['data']['attributes'][param.to_s]).not_to eq('')
            else
              expect(json['data']['attributes'][param.to_s]).to eq(get_typed_value(type))
            end
          end
        end
      end
      # main_applicant_id:
      describe 'post request with main_applicant_id' do
        let!(:applicant1) {create(:applicant, user_id: user.id)}
        
        before :each do
          authenticated_header(request, responsible)
          post 'create', {
            data: {
              attributes: {
                main_applicant_id: applicant1.id.to_s,
                
              }
            }
          }
        end
        it 'returns status code 201 and the value appears in json' do
          expect(response).to have_http_status(201)
          expect(json['data']['attributes']['main_applicant_id']).to eq(applicant1.id.to_s)
        end
      end
      
      #  applicant_ids: 'Array'
      describe 'post request with applicant_ids' do
        let!(:applicant1) {create(:applicant, user_id: user.id)}
        let!(:applicant2) {create(:applicant, user_id: user.id)}
        before :each do
          authenticated_header(request, responsible)
          post 'create', {
            data: {
              attributes: {
                main_applicant_id: applicant1.id.to_s,
                applicant_ids: [applicant1.id.to_s, applicant2.id.to_s],
              }
            }
          }
        end
        it 'returns status code 201 and the value appears in json' do
          expect(response).to have_http_status(201)
          expect(json['data']['attributes']['applicant_ids']).to eq([applicant1.id.to_s, applicant2.id.to_s])
        end
      end
      # coapplicants: 'Array'
      describe 'post request with coapplicants' do
        let!(:applicant0) {create(:applicant, user_id: user.id)}
        let!(:applicant1) {create(:applicant, user_id: user.id)}
        let!(:applicant2) {create(:applicant, user_id: user.id)}
        $request_id = nil # to store global value
        
        before :each do
          authenticated_header(request, responsible)
          post 'create', {
            data: {
              attributes: {
                main_applicant_id: applicant0.id.to_s,
                coapplicants: [{'rel_type' => 'P', 'id' => applicant1.id.to_s}, {'rel_type' => 'P', 'id' => applicant2.id.to_s}],
              }
            }
          }
        end
        
        it 'returns status code 201 and the value appears in json' do
          $request_id = json['data']['id']
          expect(response).to have_http_status(201)
          applink = Hermes::ApplicantLink.where(housing_request_id: json['data']['id'])
          expect(applink.count).to be 1
          expect(applink.first.type).to eq('P')
          expect(json['data']['attributes']['applicant_ids']).to include(applicant1.id.to_s).and(include(applicant2.id.to_s))
          expect(json['data']['attributes']['coapplicants']).to eq([{'rel_type' => 'P', 'id' => applicant1.id.to_s}, {'rel_type' => 'P', 'id' => applicant2.id.to_s}])
        end
        
        describe 'destroy previous applicants and set with new' do
          let!(:applicant3) {create(:applicant, user_id: user.id)}
          before :each do
            $request_id = json['data']['id']
            (1..2).each {|i| send("applicant#{i}").destroy}
            authenticated_header(request, responsible)
            patch 'update', {
              id: $request_id,
              data: {
                attributes: {
                  coapplicants: [
                    {'rel_type' => 'P', 'id' => applicant3.id.to_s}
                  ]
                }
              }
            }
          end
          
          it 'returns 200' do
            expect(response).to have_http_status(200)
          end
          
          it 'returns valid json' do
            expect(json['data']).to_not be_empty
            expect(json['data']['id']).to eq $request_id
            attributes = json['data']['attributes']
            expect(attributes['coapplicants'].length).to be 1
            expect(attributes['coapplicants'].first['rel_type']).to eq 'P'
            expect(attributes['coapplicants'].first['id']).to eq applicant3.id.to_s
          end
        end
        describe 'destroy previous applicants and set with new' do
          let!(:applicant3) {create(:applicant, user_id: user.id)}
          
          before :each do
            $request_id = json['data']['id']
            hr = Hermes::HousingRequest.find_by(id: json['data']['id'])
            applicant1.destroy
            applicant2.destroy
            applicant3.destroy
            
            authenticated_header(request, responsible)
            get 'show', {id: $request_id}
          end
          
          it 'returns 200' do
            expect(response).to have_http_status(200)
          end
          
          it 'returns valid json' do
            expect(json['data']).to_not be_empty
            expect(json['data']['id']).to eq $request_id
            attributes = json['data']['attributes']
            expect(attributes['coapplicants'].length).to be 0
          end
        end
      end
    end
    
    describe 'GET /housing_requests' do
      describe 'GET /housing_requests' do
        let!(:housing_request1) {create(:housing_request, documents_required: 50, main_applicant_id: applicant.id.to_s)}
        let!(:housing_request3) {create(:housing_request, documents_required: 150, main_applicant_id: applicant.id.to_s)}
        let!(:housing_request2) {create(:housing_request, documents_required: 150, main_applicant_id: applicant.id.to_s)}
        before :each do
          authenticated_header(request, responsible)
          get 'index'
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
        end
        it 'returns a valid json' do
          expect(json).not_to be_empty
          expect(json['data'].length).to eq(3)
          expect(json['meta']['pagination']).to be_present
          expect(json['meta']['pagination']['total_objects']).to eq(3)
        end
      end
      
      ACCEPTABLE_PARAMS.each do |acceptable_param|
        type = MUTATE_PARAMS[acceptable_param.to_sym] || MUTATE_PARAMS[acceptable_param.to_s]
        if type
          describe 'GET /housing_requests with sort on field ' + acceptable_param + ' works ' do
            let!(:housing_request3) {create(:housing_request, acceptable_param => get_typed_value(type, 2), main_applicant_id: applicant.id.to_s)}
            let!(:housing_request1) {create(:housing_request, acceptable_param => get_typed_value(type, 0), main_applicant_id: applicant.id.to_s)}
            let!(:housing_request2) {create(:housing_request, acceptable_param => get_typed_value(type, 1), main_applicant_id: applicant.id.to_s)}
            before :each do
              authenticated_header(request, responsible)
              get 'index', {sort: {acceptable_param => 1}}
            end
            it 'returns sorted results' do
              expect(response.status).to eq(200)
              if type == 'DateTime'
                housing_request_attribute = json['data'].map {|unit| DateTime.parse(unit['attributes'][acceptable_param])}
                expect(housing_request_attribute).to eq([get_typed_value(type, 0), get_typed_value(type, 1), get_typed_value(type, 2)])
              elsif type == 'Date'
                housing_request_attribute = json['data'].map {|unit| Date.parse(unit['attributes'][acceptable_param])}
                expect(housing_request_attribute).to eq([get_typed_value(type, 0), get_typed_value(type, 1), get_typed_value(type, 2)])
              else
                housing_request_attribute = json['data'].map {|unit| unit['attributes'][acceptable_param]}
                expect(housing_request_attribute).to eq([get_typed_value(type, 0), get_typed_value(type, 1), get_typed_value(type, 2)])
              end
            end
          end
        end
      end
      
      describe 'GET /housing_requests with operators' do
        let!(:housing_request1) {create(:housing_request, motivation1: 'acceuillant', documents_required: 50, main_applicant_id: applicant.id.to_s)}
        let!(:housing_request3) {create(:housing_request, motivation1: 'Super', documents_required: 150, main_applicant_id: applicant.id.to_s)}
        let!(:housing_request2) {create(:housing_request, motivation1: 'bel appart', documents_required: 150, main_applicant_id: applicant.id.to_s)}
        describe 'with $lt' do
          before :each do
            authenticated_header(request, responsible)
            get 'index', {documents_required: {'$lt': 100}}
          end
          it 'returns results' do
            expect(response.status).to eq(200)
            expect(json['meta']['pagination']['total_objects']).to eq(1)
          end
        end
        describe 'with $in' do
          before :each do
            authenticated_header(request, responsible)
            get 'index', {motivation1: {'$in': ['acceuillant', 'flamand']}}
          end
          it 'returns results' do
            expect(response.status).to eq(200)
            expect(json['meta']['pagination']['total_objects']).to eq(1)
          end
        end
        describe 'with $ne' do
          before :each do
            authenticated_header(request, responsible)
            get 'index', {documents_required: {'$ne': 50}}
          end
          it 'returns sorted results' do
            expect(response.status).to eq(200)
            expect(json['meta']['pagination']['total_objects']).to eq(2)
          end
        end
        describe 'with $regex' do
          before :each do
            authenticated_header(request, responsible)
            get 'index', {motivation1: {'$regex': /^S/}}
          end
          it 'returns sorted results' do
            expect(response.status).to eq(200)
            expect(json['meta']['pagination']['total_objects']).to eq(1)
          end
        end
      end
      
      describe 'GET /housing_requests with main_employer_siret' do
        let(:applicant_siret1) {create :applicant, main_employer_siret: '123456'}
        let(:applicant_siret2) {create :applicant, main_employer_siret: '987654'}
        let!(:housing_request1) {create(:housing_request, main_applicant_id: applicant.id.to_s)}
        let!(:housing_request3) {create(:housing_request, main_applicant_id: applicant_siret1.id.to_s)}
        let!(:housing_request2) {create(:housing_request, main_applicant_id: applicant_siret2.id.to_s)}
        
        before do
          authenticated_header(request, responsible)
          get 'index', {main_employer_siret: ['123456', '987654']}
        end
        it 'returns results' do
          expect(response.status).to eq(200)
          expect(json['data'].length).to eq(2)
          expect(json['data'][0]['id']).to eq housing_request3.id.to_s
          expect(json['data'][1]['id']).to eq housing_request2.id.to_s
        end
      end
    end
    
    describe 'GET /housing_requests/:id' do
      describe 'verif it work' do
        let!(:housing_request1) {create(:housing_request, motivation1: 'acceuillant', documents_required: 50, main_applicant_id: applicant.id.to_s)}
        before :each do
          authenticated_header(request, responsible)
          get 'show', {id: housing_request1.id}
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
        end
        it 'returns a valid json' do
          expect(json).not_to be_empty
        end
      end
      describe 'verif it work not work with a non authorized user' do
        let!(:housing_request1) {create(:housing_request, motivation1: 'acceuillant', documents_required: 50, main_applicant_id: applicant.id.to_s)}
        
        before :each do
          authenticated_header(request, user)
          get 'show', {id: housing_request1.id}
        end
        it 'returns status code 404' do
          expect(response).to have_http_status(404)
        end
      end
    end
    
    describe 'PATCH /housing_requests/:id' do
      housing_request_params = {}
      MUTATE_PARAMS.each do |param, type|
        housing_request_params[param] = get_typed_value(type)
      end
      
      let!(:housing_request) {create(:housing_request, **housing_request_params, main_applicant_id: applicant.id.to_s)}
      MUTATE_PARAMS.each do |param, type|
        describe 'patch request with ' + param.to_s + ' ' + type.to_s + ' is valid' do
          let(:params) do
            attr = {}
            attr[param] = get_typed_value(type, 2)
            attr
          end
          before :each do
            authenticated_header(request, responsible)
            patch 'update', {
              id: housing_request.id.to_s,
              data: {
                attributes: params
              }
            }
          end
          it 'returns status code 200 and the value appears in json' do
            expect(response).to have_http_status(200)
            if (type == 'Date' || type == 'DateTime')
              
              expect(json['data']['attributes'][param.to_s]).not_to be_empty
              expect(json['data']['attributes'][param.to_s]).not_to eq('')
            else
              expect(json['data']['attributes'][param.to_s]).to eq(get_typed_value(type, 2))
            end
          end
        end
      end
      
      describe 'DELETE /housing_requests/:id' do
        let!(:housing_request) {create(:housing_request, main_applicant_id: applicant.id.to_s, user_id: responsible.id.to_s)}
        before :each do
          authenticated_header(request, responsible)
          
          delete 'destroy', {
            id: housing_request.id.to_s
          }
        end
        it 'returns status code 404' do
          expect(response).to have_http_status(204)
        end
      end
    end
    
    
    #HOUSING REQUEST FOLDER
    describe 'GET /housing_requests/:id/housing_request_folder' do
      let!(:user1) {create(:user)}
      let!(:user2) {create(:user)}
      let!(:applicant42) {
        create(
          :applicant,
          user: user1,
          monthly_charges: {},
          nationality: "3"
        )
      }
      let!(:applicant2) {
        create(
          :applicant,
          user: user2,
          monthly_charges: {},
          nationality: "2"
        )
      }
      let!(:housing_request5) {
        create(
          :housing_request,
          is_active: true,
          main_applicant: applicant42,
          applicant_ids: [applicant2.id.to_s],
          dependents: [{
                         "last_name" => "GOUBILL",
                         "first_name" => "CHRIS",
                         "birth_date" => "2017-03-23T00:00:00.000Z",
                         "rel_type" => "E",
                         "coparentality" => "G",
                         "sex" => "M"},
                       {"last_name" => "GOUBILL",
                        "first_name" => "CHRIsdS",
                        "birth_date" => "1990-03-23T00:00:00.000Z",
                        "rel_type" => "E",
                        "coparentality" => "G",
                        "sex" => "M"},
                       {"last_name" => "GOUBILLze",
                        "first_name" => "CHRIS",
                        "birth_date" => "2017-03-23T00:00:00.000Z",
                        "rel_type" => "P",
                        "sex" => "M"}]
        )
        
      }
      describe 'get ok' do
        before :each do
          file = File.open([Hermes::Engine.root, %w[lib hermes tools minion.jpg]].flatten.join('/'))
          blob = file.read
          
          RequiredDocument.all.each do |required_doc|
            required_doc.set_blobs([{blob: blob, filename: "minion.jpg"}])
            required_doc.save
          end
          authenticated_header(request, responsible)
          
          get 'housing_request_folder', {
            id: housing_request5.id.to_s
          }
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
        end
        it 'returns a valid json' do
          expect(json).not_to be_empty
          expect(json["data"]["attributes"]["required_documents"].length).to eq(6)
          expect(json["data"]["attributes"]["nud_status"]).to eq("Dossier complet prêt pour traitement")
        end
      end
    
      describe 'start_treatment' do
        before :each do
          file = File.open([Hermes::Engine.root, %w[lib hermes tools minion.jpg]].flatten.join('/'))
          blob = file.read
    
          RequiredDocument.all.each do |required_doc|
            required_doc.set_blobs([{blob: blob, filename: "minion.jpg"}])
            required_doc.save
    
          end
          authenticated_header(request, responsible)
    
          post 'start_treatment', {
            id: housing_request5.id.to_s
          }
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
         expect(housing_request5.housing_request_folder.reload.nud_status).to eq("Dossier validé")
        end
      end
    end
    
    context 'as reservataire' do
      let!(:reservation_holder1) {create(:reservation_holder)}
      let!(:reservation_holder2) {create(:reservation_holder)}
      let!(:user_applicant) {create(:user, auth_roles: ['Demandeur'])}
      let!(:user1) {create(:user, auth_roles: ['Réservataire'], reservation_holder_id: reservation_holder1.id)}
      let!(:user2) {create(:user, auth_roles: ['Réservataire'], reservation_holder_id: reservation_holder2.id)}
      let!(:user3) {create(:user, auth_roles: ['Demandeur'])}
      let!(:main_applicant) {create(:applicant, user: user_applicant)}
      let!(:housing_request1) {create(:housing_request, main_applicant_id: main_applicant.id, reservation_holder_id: reservation_holder1.id)}
      let!(:housing_request2) {create(:housing_request, main_applicant_id: main_applicant.id, reservation_holder_id: reservation_holder2.id)}
      let!(:housing_request3) {create(:housing_request, main_applicant_id: main_applicant.id)}
      let!(:housing_request4) {create(:housing_request, main_applicant_id: main_applicant.id)}
      
      describe 'GET /housing_requests' do
        describe 'access correct request' do
          before :each do
            authenticated_header(request, user1)
            get 'index'
          end
          it 'returns results' do
            expect(response.status).to eq(200)
            expect(json['meta']['pagination']['total_objects']).to eq(1)
          end
        end
      end
      
      describe 'GET /housing_requests/:id' do
        describe 'has access' do
          before :each do
            authenticated_header(request, user1)
            get 'show', {id: housing_request1.id}
          end
          it 'returns results' do
            expect(response.status).to eq(200)
          end
        end
        describe 'is forbidden to access' do
          before :each do
            authenticated_header(request, user1)
            get 'show', {id: housing_request2.id}
          end
          it 'returns results' do
            expect(response.status).to eq(404)
          end
        end
      end
      
      describe 'PATCH /housing_requests/:id' do
        describe 'has access' do
          before :each do
            authenticated_header(request, user1)
            patch 'update', {
              id: housing_request1.id.to_s,
              data: {
                attributes: {
                  contract_type: "test"
                }
              }
            }
          end
          it 'returns results' do
            expect(response.status).to eq(200)
            expect(housing_request1.reload.contract_type).to eq('test')
          end
        end
        
        describe 'is forbidden to access' do
          before :each do
            authenticated_header(request, user1)
            patch 'update', {
              id: housing_request2.id.to_s,
              data: {
                attributes: {
                  contract_type: 'Type'
                }
              }
            }
          end
          it 'returns results' do
            expect(response.status).to eq(404)
          end
        end
      end
      
      describe 'DELETE /housing_requests/:id' do
        describe 'has access' do
          before :each do
            authenticated_header(request, user1)
            delete 'destroy', {
              id: housing_request1.id.to_s
            }
          end
          it 'returns results' do
            expect(response.status).to eq(204)
          end
        end
        describe 'is forbidden to access' do
          before :each do
            authenticated_header(request, user1)
            delete 'destroy', {
              id: housing_request2.id.to_s
            }
          end
          it 'returns results' do
            expect(response.status).to eq(404)
          end
        end
      end
    end

    context 'desactivate and delete' do
      let!(:admin) {create(:user, roles: ['Administrator'])}
      let!(:user_applicant) {create(:user, auth_roles: ['Demandeur'])}
      let!(:main_applicant) {create(:applicant, user: user_applicant)}
      let!(:housing_request1) {create(:housing_request, user: admin, main_applicant_id: main_applicant.id, is_active: true)}
      let!(:housing_request2) {create(:housing_request, user: admin, main_applicant_id: main_applicant.id, is_active: true)}
      let!(:housing_offer1) {create(:housing_offer)}
      let!(:housing_offer2) {create(:housing_offer)}
      let!(:housing_offer3) {create(:housing_offer)}
      let!(:housing_application1) {create(:housing_application, housing_request: housing_request1, housing_offer: housing_offer1)}
      let!(:housing_application2) {create(:housing_application, housing_request: housing_request1, housing_offer: housing_offer2)}
      
      describe 'PATCH /housing_requests/:id' do
        describe 'on request with applications' do
          before :each do
            authenticated_header(request, admin)
            patch 'update', {
              id: housing_request1.id.to_s,
              data: {
                attributes: {
                  is_active: false
                }
              }
            }
          end
          it 'returns results' do
            expect(housing_request1.housing_offer_ids).to eq([housing_offer1.id.to_s, housing_offer2.id.to_s])
            expect(response.status).to eq(422)
          end
        end

        describe 'on request without applications' do
          before :each do
            authenticated_header(request, admin)
            patch 'update', {
              id: housing_request2.id.to_s,
              data: {
                attributes: {
                  is_active: false
                }
              }
            }
          end
          it 'returns results' do
            expect(response.status).to eq(200)
          end
        end
      end

      describe 'DELETE /housing_requests/:id' do
        describe 'on request with applications' do
          before :each do
            authenticated_header(request, admin)
            delete 'destroy', {
              id: housing_request1.id.to_s
            }
          end
          it 'returns results' do
            expect(response.status).to eq(422)
          end
        end
        describe 'on request without applications' do
          before :each do
            authenticated_header(request, admin)
            delete 'destroy', {
              id: housing_request2.id.to_s
            }
          end
          it 'returns results' do
            expect(response.status).to eq(204)
          end
        end
      end
    end
  end
  
  require_relative '../../swagger_helper'
  
  describe Api::Dmo::HousingRequestsController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
    TAGS_API_DMO_HOUSING_REQUESTS = 'API_DMO_HOUSING_REQUESTS'
    include JwtAuthHelper
    extend HermesSwaggerJsonParams
    include SwaggerJsonAttributes
    let!(:responsible) {create(:user, first_name: 'Martin', last_name: 'Durand', email: 'martin.durand@gmail.com', roles: ['Administrator'])}
    let!(:user) {create(:user, first_name: 'Jackie', last_name: 'Mike', email: 'jackie.mike@gmail.com')}
    let!(:applicant) {create(:applicant, user_id: responsible.id.to_s, nationality: "1")}
    context 'as admin' do
      let!(:admin) {create(:administrator)}
      
      path '/api/dmo/housing_requests' do
        post 'Create housing_request' do
          tags TAGS_API_DMO_HOUSING_REQUESTS
          produces 'application/json'
          consumes 'application/json'
          security [Bearer: {}]
          json_params = {}
          json_params[:main_applicant_id] = {type: :string}
          json_params[:coapplicants] = {type: :array}
          json_params[:applicant_ids] = {type: :array}
          json_params[:reference_tax_revenues] = {type: :float}
          json_params[:dependents] = {type: :array}
          MUTATE_PARAMS.each do |key, val|
            json_params[key] = {type: nice_type(val)}
          end
          parameter name: :params, in: :body, schema: jsonify_params(json_params)
          
          response '201', 'housing_request created' do
            let(:Authorization) {"Bearer #{token_for(responsible)}"}
            let(:params) do
              jsonapi_attributes(
                {main_applicant_id: applicant.id.to_s}
              )
            end
            include_context 'with integration test'
            include_context 'valid jsonapi'
            include_context 'not containings errors'
            
            after do |_example|
              json = JSON.parse(response.body)
              
              expect(json['data']).to be_a Hash
              expect(json['data']['id']).to_not be nil
              expect(json['data']['type']).to eq 'housing_requests'
            end
          end
        end
      end
      
      path '/api/dmo/housing_requests' do
        get 'Retrieves all housing_requests' do
          tags TAGS_API_DMO_HOUSING_REQUESTS
          produces 'application/json'
          
          parameter name: 'ids[]', in: :query, type: :array, items: {type: :string}, collectionFormat: :multi, required: false
          ACCEPTABLE_PARAMS.each do |acceptable_param|
            type = :string
            if MUTATE_PARAMS[acceptable_param]
              type = nice_type(MUTATE_PARAMS[acceptable_param])
            end
            parameter name: acceptable_param.to_s, in: :query, type: type, required: false
          end
          # VALID_OPERATORS.each do |valid_operator|
          #    parameter( name: valid_operator.to_s, in: :query,type: :object, required: false)
          #  end
          
          security [Bearer: {}]
          paginator
          
          response '200', 'housing_requests found' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let!(:housing_requests) {create_list(:housing_request, 5, main_applicant_id: applicant.id.to_s)}
            
            include_context 'with integration test'
            include_context 'valid jsonapi collection' do
              let(:expected_collection) {housing_requests}
            end
          end
        end
      end
      
      path '/api/dmo/housing_requests/{id}' do
        get 'Retrieves a single housing_request by id' do
          tags TAGS_API_DMO_HOUSING_REQUESTS
          produces 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]
          
          response '200', 'housing_request found' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:housing_request) {create(:housing_request, main_applicant_id: applicant.id.to_s)}
            let(:id) {housing_request.id}
            
            include_context 'with integration test'
            include_context 'valid jsonapi ressource' do
              let(:expected_ressource) {housing_request}
            end
          end
        end
      end
      
      path '/api/dmo/housing_requests/clp_details' do
        get 'Retrieves a single housing_request by id' do
          tags TAGS_API_DMO_HOUSING_REQUESTS
          produces 'application/json'
          parameter name: 'als_ids[]', in: :query, type: :array, items: {type: :string}, required: true, collectionFormat: :multi
          ACCEPTABLE_PARAMS.each do |acceptable_param|
            type = :string
            if MUTATE_PARAMS[acceptable_param]
              type = nice_type(MUTATE_PARAMS[acceptable_param])
            end
            parameter name: acceptable_param.to_s, in: :query, type: type, required: false
          end
          security [Bearer: {}]
          paginator
          response '200', 'housing_request found' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:housing_requests) {create_list(:housing_request, 5, main_applicant_id: applicant.id.to_s)}
            let(:requests_sample) {2.times.map {|i| housing_requests[i]}}
            let!(:'als_ids[]') {requests_sample.map(&:als_id)}
            
            include_context 'with integration test'
            include_context 'valid jsonapi collection' do
              let(:expected_collection) {requests_sample}
            end
          end
        end
      end
      
      path '/api/dmo/housing_requests/{id}/housing_request_folder' do
        get 'Retrieves the housing_request_folder of an housing_request' do
          tags TAGS_API_DMO_HOUSING_REQUESTS
          produces 'application/json'
          parameter name: :id, in: :path, type: :string, required: true
          security [Bearer: {}]
          response '200', 'housing_request_folder found' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:housing_request) {create(:housing_request, main_applicant_id: applicant.id.to_s)}
            let(:id) {housing_request.id}
            
            include_context 'with integration test'
            include_context 'valid jsonapi'
            include_context 'not containings errors'
            after do |_example|
              JSON.parse(response.body)
            end
          end
        end
      end
      
      path '/api/dmo/housing_requests/{id}' do
        put 'edit housing_request' do
          tags TAGS_API_DMO_HOUSING_REQUESTS
          produces 'application/json'
          consumes 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]
          json_params = {}
          MUTATE_PARAMS.each do |key, val|
            json_params[key] = {type: nice_type(val)}
          end
          json_params[:main_applicant_id] = {type: :string}
          json_params[:coapplicants] = {type: :array}
          json_params[:applicant_ids] = {type: :array}
          json_params[:reference_tax_revenues] = {type: :float}
          json_params[:dependents] = {type: :array}
          
          parameter name: :params, in: :body, schema: jsonify_params(json_params)
          #jsonapi_params( **json_params )
          
          response '200', 'housing_request updated' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:housing_request) {create(:housing_request, main_applicant_id: applicant.id.to_s)}
            let(:id) {housing_request.id}
            let(:params) do
              jsonapi_attributes(
                {
                  attributed_housing_district: "ACCEPTEE EN CAL"
                }
              )
            end
            include_context 'with integration test'
            include_context 'valid jsonapi'
            include_context 'not containings errors'
            
            after do |_example|
              json = JSON.parse(response.body)
              
              expect(json['data']).to be_a Hash
              expect(json['data']['id']).to_not be nil
              expect(json['data']['type']).to eq 'housing_requests'
            end
          end
        end
      end
      
      path '/api/dmo/housing_requests/{id}' do
        delete 'Delete a single housing_request by id' do
          tags TAGS_API_DMO_HOUSING_REQUESTS
          produces 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]
          
          response '204', 'housing_request deleted' do
            let(:Authorization) {"Bearer #{token_for(responsible)}"}
            let(:id) {create(:housing_request, user_id: responsible.id.to_s, main_applicant_id: applicant.id.to_s).id}
            run_test!
          end
        end
      end
    end
  end
end
